"""The 'worktoy.yolo' package provides tools for testing functions and
running tests."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._yolo import yolo
from ._run_tests import runTests
