"""The 'worktoy.keenum' package provides the KeeNum class for enumeration!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._keenum_object import KeeNumObject
from ._keenum_space import KeeNumSpace, auto
from ._keenumeta import KeeNumMeta, KeeNum
